import argparse

from statsmodels.iolib.smpickle import load_pickle

import mksc
from custom import Custom
from mksc.feature_engineering import preprocess
from mksc.feature_engineering import transform
from mksc.model import training


def main(custom=False, **kwargs):
    """
    模型训练主程序入口
    """
    feature_engineering = load_pickle('result/feature_engineering.pickle')
    data = mksc.load_data()
    numeric_var, category_var, datetime_var, label_var = preprocess.get_variable_type()
    feature = data[numeric_var + category_var + datetime_var]
    label = data[label_var]

    cs = Custom()
    # 自定义数据清洗
    feature, label = cs.clean_data(feature, label)

    # 数据类型转换
    feature[numeric_var] = feature[numeric_var].astype('float')
    feature[category_var] = feature[category_var].astype('object')
    feature[datetime_var] = feature[datetime_var].astype('datetime64')

    # 自定义特征组合模块
    feature = cs.feature_combination(feature)

    # 数据处理
    feature = transform(feature, feature_engineering)
    feature = feature[feature_engineering['feature_selected']]

    # 模型训练,custom=True选择自定义模型
    if custom:
        training(feature, label, use=cs.model(), **kwargs)
    else:
        training(feature, label, **kwargs)


if __name__ == "__main__":
    args = argparse.ArgumentParser()
    args.add_argument("-m", "--model", type=str, help="模型选择，默认自动选择最优")
    args.add_argument("-r", "--resample", type=bool, default=False, help="是否使用重采样，默认不采样")
    args.add_argument("-c", "--custom", type=bool, default=False, help="是否使用自定义模型，默认不采样")
    accepted = vars(args.parse_args())
    main(model_name=accepted['model'], resample=accepted['resample'], custom=accepted['custom'])



import argparse

import mksc
from custom import Custom
from mksc.feature_engineering import FeatureEngineering
from mksc.feature_engineering import preprocess


def main(adjust=False, **kwargs):
    """
    项目特征工程程序入口
    """
    # 加载数据、变量类型划分、特征集与标签列划分
    print(" >>> 数据加载...")
    data = mksc.load_data()
    numeric_var, category_var, datetime_var, label_var = preprocess.get_variable_type()
    feature = data[numeric_var + category_var + datetime_var]
    label = data[label_var]
    print(f" >>> 当前数据规模: {data.shape}\n"
          f"     -- 数值型特征：{numeric_var}\n"
          f"     -- 类别型特征：{category_var}\n"
          f"     -- 日期型特征：{datetime_var}\n"
          f"     待筛选特征数：{feature.shape[1]}\n"
          )

    cs = Custom()
    # 自定义数据清洗
    feature, label = cs.clean_data(feature, label)

    # 数据类型转换
    feature[numeric_var] = feature[numeric_var].astype('float')
    feature[category_var] = feature[category_var].astype('object')
    feature[datetime_var] = feature[datetime_var].astype('datetime64')

    # 自定义特征组合，全部为数值变量
    feature = cs.feature_combination(feature)

    # 调整变量，只能减少
    if adjust:
        feature = cs.feature_adjust(feature)
    cs.show()

    # 标准化特征工程
    print(">>> 启动特征工程标准进程")
    fe = FeatureEngineering(feature, label, **kwargs)
    fe.run()


if __name__ == "__main__":
    args = argparse.ArgumentParser()
    args.add_argument("-m", "--missing_threshold", type=tuple, default=(0.95, 0.05), help="缺失值阈值,默认(0.95, 0.05)")
    args.add_argument("-d", "--distinct_threshold", type=float, default=0.95, help="唯一率阈值,默认0.95")
    args.add_argument("-u", "--unique_threshold", type=float, default=0.95, help="众数阈值,默认0.95")
    args.add_argument("-a", "--abnormal_threshold", type=float, default=0.05, help="极端值阈值,默认0.05")
    args.add_argument("-c", "--correlation_threshold", type=float, default=0.7, help="相关系数阈值,默认0.7")
    args.add_argument("-v", "--variance_threshold", type=float, default=0.05, help="方差阈值,默认0.05")
    args.add_argument("-s", "--stepwise", type=bool, default=True, help="是否启用逐步回归筛选")
    accepted = vars(args.parse_args())
    missing_threshold = accepted.get("missing_threshold")
    distinct_threshold = accepted.get("distinct_threshold")
    unique_threshold = accepted.get("unique_threshold")
    abnormal_threshold = accepted.get("abnormal_threshold")
    correlation_threshold = accepted.get("correlation_threshold")
    variance_threshold = accepted.get("variance_threshold")
    stepwise = accepted.get("stepwise")
    main(missing_threshold=missing_threshold, distinct_threshold=distinct_threshold,
         unique_threshold=unique_threshold, abnormal_threshold=abnormal_threshold,
         correlation_threshold=correlation_threshold, variance_threshold=variance_threshold,
         stepwise=stepwise)



import argparse
import configparser
import os
from datetime import date
from math import log

import pandas as pd
from statsmodels.iolib.smpickle import load_pickle

import mksc
from custom import Custom
from mksc.feature_engineering import preprocess
from mksc.feature_engineering import scoring
from mksc.feature_engineering import transform


def main(model_name, dataset="apply", card=False, score=False, remote=False, local=False):
    # 数据、模型加载
    model, threshold = load_pickle(f'result/{model_name}.pickle')

    feature_engineering = load_pickle('result/feature_engineering.pickle')
    data = mksc.load_data(dataset, local=local)
    numeric_var, category_var, datetime_var, label_var = preprocess.get_variable_type()
    numeric_var, category_var, datetime_var = [list(set(t) & set(data.columns)) for t in (numeric_var, category_var, datetime_var)]
    feature = data[numeric_var + category_var + datetime_var]
    label = []

    cs = Custom()
    # 自定义数据清洗
    feature, label = cs.clean_data(feature, label)

    # 数据类型转换
    feature[numeric_var] = feature[numeric_var].astype('float')
    feature[category_var] = feature[category_var].astype('object')
    feature[datetime_var] = feature[datetime_var].astype('datetime64')

    # 自定义特征组合模块
    feature = cs.feature_combination(feature)

    feature = feature[feature_engineering['feature_selected']]

    # 数据处理
    feature = transform(feature, feature_engineering)

    # 应用预测
    print(">>> 应用预测")
    res_label = pd.DataFrame(model.predict(feature), columns=['label_predict'])
    res_prob = pd.DataFrame(model.predict_proba(feature), columns=['probability_0', "probability_1"])
    res_prob['res_odds'] = res_prob['probability_0'] / res_prob["probability_1"]
    res_prob['label_threshold'] = res_prob['probability_1'].apply(lambda x: 0 if x < threshold else 1)
    res = pd.concat([data, res_label, res_prob], axis=1)

    if card and model_name == 'lr':
        # 转化评分
        score_card = load_pickle('result/card.pickle')
        res = scoring.transform_score(res, score_card)

    if score:
        print(">>> 概率转换评分")
        cfg = configparser.ConfigParser()
        cfg.read(os.path.join(os.getcwd(), 'template/config', '——configuration.ini'), encoding='utf_8_sig')
        odds = cfg.get('SCORECARD', 'odds')
        score = cfg.get('SCORECARD', 'score')
        pdo = cfg.get('SCORECARD', 'pdo')
        a, b = scoring.make_score(odds, score, pdo)
        res['score'] = res_prob['res_odds'].apply(lambda x: a + b * log(float(x)))
        if dataset == "all":
            if cs.adjust_bins:
                bins = cs.adjust_bins
            else:
                from mksc.feature_engineering.binning import tree_binning
                bins = tree_binning(res[label_var], res['score'].to_frame())[0]["result"]["score"]
            print(">>> 训练集分组")
            res['level'] = pd.cut(res['score'], bins)
            temp = res.groupby("level", as_index=False).count()
            temp['rate'] = temp[label_var] / feature.shape[0]
            temp = temp[['level', 'rate']]
            print(temp)
            print(res.head())

    # 结果保存
    print(f">>> 结果保存中，保存模式：{remote}")
    res['load_date'] = str(date.today())
    mksc.save_result(res, filename="apply_result.csv", remote=remote)


if __name__ == '__main__':
    args = argparse.ArgumentParser()
    args.add_argument("-m", "--model", type=str, required=True, help=r"模型选择：xgb\lr\svm\rf\gbdt\lgb\dt\自定义")
    args.add_argument("-d", "--dataset", type=str, default="apply", help="预测数据集")
    args.add_argument("-c", "--card", type=bool, default=False, help="是否制作评分卡")
    args.add_argument("-s", "--score", type=bool, default=False, help="是否转换成评分")
    args.add_argument("-r", "--remote", type=bool, default=False,  help="是否保存远程")
    args.add_argument("-l", "--local", type=bool, default=False,  help="是否读取远程")
    accepted = vars(args.parse_args())
    main(model_name=accepted['model'], dataset=accepted['dataset'], card=accepted['card'],
         score=accepted['score'], remote=accepted['remote'], local=accepted['local'])



import argparse

import pandas as pd
import pandas_profiling as pp
from statsmodels.iolib.smpickle import save_pickle

import mksc


def main(report=False, apply=False, local=False):
    """
    探索性数据分析主程序入口
    Args:
        report: 是否保留分析报告，由于该报告对大数据集会占用大量时间，默认不生成
        apply: 是否读取应用集并保存pickle文件，此处注意应用集规模，默认生成
        local: 是否读取本地文件，默认读取
    生成以下四份文件：
        1、数据对象文件：         “data/data.pickle”
        2、变量类型配置文件：      “config/variable_type.csv”
        3、样例数据：             “data/sample.xlsx”
        4、结果报告：             “result/report.html”
    """
    # 加载数据并保存本地
    print(">>> 全量数据集加载...")
    data = mksc.load_data(mode="all", local=local)
    save_pickle(data, 'data/data.pickle')

    # 生成变量类别配置文件
    print(">>> 生成配置表config/variable_type.csv，请完善配置...")
    # 变量是否保留进行特征工程(isSave): 0-不保留；1-保留
    # 变量类型(Type): numeric-数值类型；category-类别类型；datetime-日期类型；label-标签列；identifier-业务标识符
    res = pd.DataFrame(zip(data.columns, [1]*len(data.columns), ['numeric']*len(data.columns), ['']*len(data.columns)),
                       columns=['Variable', 'isSave:[0/1]', 'Type:[identifier/numeric/category/datetime/label]', 'Comment'])
    res.to_csv("config/variable_type.csv", index=False)

    # 抽样探索
    print(">>> 生成探索性分析抽样data/sample.xlsx")
    sample = data.sample(min(len(data), 1000))
    sample.reset_index(drop=True, inplace=True)
    with pd.ExcelWriter('data/sample.xlsx') as writer:
        sample.to_excel(writer, sheet_name="抽样数据1000条", index=False)
        sample.corr().to_excel(writer, sheet_name="相关系数", index=False)
        sample.select_dtypes(exclude=['object', 'datetime']).describe().to_excel(writer, sheet_name="数值数据汇总")
        sample.select_dtypes(include=['object', 'datetime']).describe().to_excel(writer, sheet_name="分类数据汇总")

    # 保存分析报告
    if report and (not sample.empty):
        print(">>> 生成数据分析报告result/report.html")
        report = pp.ProfileReport(sample)
        report.to_file('result/report.html')

    # 保存应用集，注意规模
    if apply:
        print(">>> 保存应用集到本地")
        apply = mksc.load_data(mode="apply", local=local)
        save_pickle(apply, 'data/apply.pickle')


if __name__ == '__main__':
    args = argparse.ArgumentParser()
    args.add_argument("-r", "--report", type=bool, default=False, help="是否生成报告,默认不生成")
    args.add_argument("-a", "--apply", type=bool, default=False, help="是否保存应用集二进制文件,默认不生成")
    args.add_argument("-l", "--local", type=bool, default=False, help="是否读取本地文件,默认不读取")
    accepted = vars(args.parse_args())
    main(report=accepted.get("report"), apply=accepted.get("apply"), local=accepted.get("local"))



import pickle

from mksc.feature_engineering import binning
from mksc.feature_engineering import seletction
from mksc.feature_engineering import values


class FeatureEngineering(object):

    def __init__(self, feature, label,
                 missing_threshold=(0.95, 0.05), distinct_threshold=0.95, unique_threshold=0.95,
                 abnormal_threshold=0.05, correlation_threshold=0.7, variance_threshold=0.05, **kwargs):
        self.feature = feature
        self.label = label
        self.missing_threshold = missing_threshold
        self.distinct_threshold = distinct_threshold
        self.unique_threshold = unique_threshold
        self.abnormal_threshold = abnormal_threshold
        self.correlation_threshold = correlation_threshold
        self.variance_threshold = variance_threshold
        self.threshold = {"missing_threshold": self.missing_threshold,
                          "distinct_threshold": self.distinct_threshold,
                          "unique_threshold": self.unique_threshold,
                          "abnormal_threshold": self.abnormal_threshold,
                          "correlation_threshold": self.correlation_threshold,
                          "variance_threshold": self.variance_threshold}
        self.kwargs = kwargs
        print(self.threshold)
        print(self.kwargs)

    def run(self):
        """
        特征工程过程函数,阈值参数可以自定义修改
        1. 基于统计特性特征选择：缺失率、唯一率、众数比例
        2. 缺失值处理
        TODO 异常值处理
        4. 极端值处理
        5. 正态化处理
        6. 归一化处理
        7. 最优分箱
        8. IV筛选
        TODO PSI筛选
        9. 相关性筛选
        10. woe转化
        11. One-Hot
        12. TODO 降维

        Returns:
            feature: 已完成特征工程的数据框
            label: 已完成特征工程的标签列
        """
        feature = self.feature.copy()
        label = self.label.copy()

        # 单变量筛选：基于缺失率、唯一率、众数比例统计特征筛选
        print(">>> 单变量Filter: 缺失值过滤")
        missing_value = seletction.get_missing_value(feature, self.missing_threshold[0])
        print(f"    -- {missing_value['drop']}")
        print(">>> 单变量Filter: 唯一率过滤")
        distinct_value = seletction.get_distinct_value(feature, self.distinct_threshold)
        print(f"    -- {distinct_value['drop']}")
        print(">>> 单变量Filter: 众数比率过滤")
        unique_value = seletction.get_unique_value(feature, self.unique_threshold)
        print(f"    -- {unique_value['drop']}")
        feature.drop(set(missing_value['drop'] + distinct_value['drop'] + unique_value['drop']), axis=1, inplace=True)

        # 缺失值处理
        print(">>> 单变量预处理: 缺失值补充")
        feature, missing_filling = values.fix_missing_value(feature, self.missing_threshold[1])

        # 极端值处理
        print(">>> 单变量预处理: 极端值值修正")
        feature, abnormal_value = values.fix_abnormal_value(feature,
                                                            self.abnormal_threshold,
                                                            method=self.kwargs.get("method", "boundary"))

        # 标准化处理
        print(">>> 单变量预处理: 标准归一化")
        feature, scale_result = values.fix_scaling(feature, self.variance_threshold)
        feature.drop(scale_result['drop'], axis=1, inplace=True)

        # 正态化处理
        standard_lambda = None
        if self.kwargs.get("standard", False):
            feature, standard_lambda = values.fix_standard(feature)

        # 数值特征最优分箱，未处理的变量，暂时退出模型
        print(">>> 单变量预处理：最优分箱")
        bin_result, iv_result, woe_result, woe_adjust_result = binning.tree_binning(label, feature)
        bin_error_drop = bin_result['error'] + woe_adjust_result
        print(f"    -- 空值规则预测：{bin_result['error']}")
        print(f"    -- 非空值规则预测：{woe_adjust_result}")
        print(f"    -- {bin_error_drop}")

        # IV筛选
        print(">>> 单变量Filter: IV值过滤")
        iv_drop = list(filter(lambda x: iv_result[x] < 0.02, iv_result))
        print(f"    -- {iv_drop}")
        feature.drop(iv_drop + bin_error_drop, inplace=True, axis=1)

        # 相关性筛选
        print(">>> 多变量Filter: 相关系数过滤")
        cor_drop = seletction.get_cor_drop(feature, iv_result, self.correlation_threshold)
        print(f"    -- {cor_drop}")
        feature.drop(cor_drop, inplace=True, axis=1)

        # woe转化
        print(">>> 单变量预处理: WOE转换")
        feature = binning.woe_transform(feature, woe_result, bin_result)

        # 逐步回归
        if self.kwargs.get("stepwise", False):
            print(">>> 多变量Filter: 逐步回归")
            feature_selected = seletction.model.stepwise_selection(feature, label)
        else:
            # TODO 降维
            feature_selected = feature.columns

        # 中间结果保存
        result = {"missing_value": missing_value,
                  "distinct_value": distinct_value,
                  "unique_value": unique_value,
                  "abnormal_value": abnormal_value,
                  "missing_filling": missing_filling,
                  'standard_lambda': standard_lambda,
                  'scale_result': scale_result,
                  "bin_result": bin_result,
                  "iv_result": iv_result,
                  "woe_result": woe_result,
                  "woe_adjust_result": woe_adjust_result,
                  "bin_error_drop": bin_error_drop,
                  "iv_drop": iv_drop,
                  "cor_drop": cor_drop,
                  "feature_selected": feature_selected
                  }
        print(">>> 特征工程简单报告:")
        print(f"    缺失值过滤列数量:{missing_value['drop_number']}")
        print(f"    唯一值过滤列数量:{distinct_value['drop_number']}")
        print(f"    众数值过滤列数量:{unique_value['drop_number']}")
        print(f"    IV值过滤列数量:{len(iv_drop)}")
        print(f"    相关性过滤列数量:{len(cor_drop)}")
        print(f"    保留特征:{feature_selected}")
        print(f"    保留数量:{len(feature_selected)}")

        with open('result/feature_engineering.pickle', 'wb') as f:
            f.write(pickle.dumps(result))

        with open('result/apply_sql.txt', 'w') as f:
            feature_selected = set([v.split("__")[0] for v in result["feature_selected"]])
            sql = ",".join(feature_selected)
            f.write(f"你可以使用以下取数语句避免应用集过大无法进行select *操作，注意提前剔除组合变量\n\nselect {sql} from ")



from .abnormal import fix_abnormal_value
from .missing import fix_missing_value
from .normalization import normalization
from .scale import fix_scaling
from .standard import fix_standard, logarithmetics



def normalization(feature):
    """
    对数据框中的数据进行最大最小值标准化

    Args:
        feature: 待处理的数据框
    Returns:
        feature: 已处理数据框
        normalization_result: 归一化结果集
    """
    numeric_var = feature.select_dtypes(exclude=['object', 'datetime']).columns
    normalization_result = {"result": {}}
    for c in numeric_var:
        sm = feature[c].describe()
        normalization_result[c] = {'max': sm['max'], 'min': sm['min']}
        feature[c] = feature[c].apply(lambda x: (x - sm['min'])/(sm['max']-sm['min']) if x else x)
    return feature, normalization_result



import os

from mksc.core import reader


def save_result(data, filename=None, remote=False):
    cfg = reader.config()
    if remote:
        data.to_sql(cfg.get('DATABASE', 'SCHEMA_NAME'),
                    cfg.get('DATABASE', 'SAVE_ENGINE_URL'),
                    if_exists='append',
                    index=False,
                    chunksize=5000)
    data.to_csv(os.path.join(os.getcwd(), 'result', filename), index=False)



import configparser
import os

import pandas as pd
from statsmodels.iolib.smpickle import load_pickle


def file(filename):
    if '.csv' in filename:
        data = pd.read_csv(filename)
        return data
    elif '.pickle' in filename:
        data = load_pickle(filename)
        return data
    elif ('.xls' in filename) or ('.xlsx' in filename):
        data = pd.read_excel(filename)
        return data
    else:
        print("Wrong Data Type")

def config():
    """
    读取项目配置文件/config/configuration.ini
    Returns:
        cfg: 配置参数对象
    """
    cfg = configparser.ConfigParser()
    cfg.read(os.path.join(os.getcwd(), 'config', 'configuration.ini'), encoding='utf_8_sig')
    return cfg



import pandas as pd


class Custom(object):
    """
    自定义预处理类函数封装
    """
    def __init__(self):
        self.cleaned_var = []
        self.used_var = []
        self.adjust_var = []
        self.adjust_bins = []

    def clean_data(self, feature, label):
        """
        基于单元格与行操作的数据清洗

        Args:
            feature: 待清洗特征数据框
            label: 待清洗标签序列

        Returns:
            feature_tmp: 已清洗特征数据框
            label_tmp: 已清洗标签序列
        """
        feature_tmp = feature.copy()
        label_tmp = label.copy()
        # ------------------------------------------
        # 自定义清洗, 每次使用均需保存变量名
        # eg.
        #     feature_tmp['variable'] = feature_tmp['variable']
        #     self.cleaned_var.append('old_variable')
        # ------------------------------------------
        return feature_tmp, label_tmp

    def feature_combination(self, feature):
        """
        基于列操作的数据清洗与特征构造

        Args:
            feature: 待清洗特征数据框

        Returns:
            feature_tmp: 已清洗特征数据框
        """
        feature_tmp = feature.copy()
        # ------------------------------------------
        # 构造衍生变量, 每次使用均需保存变量名
        # eg: feature_tmp['new_variable'] = feature_tmp['old_variable']
        #     feature_tmp.drop('old_variable', axis=1, inplace=True)
        #     self.used_var.append('old_variable')
        # ------------------------------------------

        # 日期变量处理, 提取完后需要丢弃变量
        datetime_var = feature_tmp.select_dtypes(include='datetime64[ns]').columns
        date_var = ['day', 'dayofweek', 'dayofyear', 'days_in_month', 'is_leap_year', 'is_month_end',
                    'is_month_start', 'is_quarter_end', 'is_quarter_start', 'is_year_end',
                    'is_year_start', 'month', 'quarter', 'week', 'weekday', 'weekofyear', 'year']
        for dv in datetime_var:
            for v in date_var:
                feature_tmp[f"{dv}__{v}"] = eval(f"feature_tmp['{dv}'].dt.{v}")
                feature_tmp[f"{dv}__{v}"] = feature_tmp[f"{dv}__{v}"].astype("object")
        feature_tmp.drop(datetime_var, axis=1, inplace=True)

        # 分类变量one-hot处理，处理完成后需要丢弃变量
        category_var = feature_tmp.select_dtypes(include=['object']).columns
        if not category_var.empty:
            feature_tmp[category_var].fillna("NA", inplace=True)
            one_hot = pd.get_dummies(feature_tmp[category_var], prefix_sep="__")
            one_hot.columns = list(map(lambda x: x.split(".")[0], one_hot.columns))
            feature_tmp = pd.concat([feature_tmp, one_hot], axis=1)
        feature_tmp.drop(category_var, axis=1, inplace=True)
        return feature_tmp

    def feature_adjust(self, feature):
        """
        调整特征表
        Args:
            feature: 原始特征表

        Returns:
            feature_tmp: 调整后的特征数据框
        """
        # ------------------------------------------
        # 后期调整的时候控制某些特征不进入训练，只能减少特征
        # ------------------------------------------
        self.adjust_var = []
        feature_tmp = feature[self.adjust_var]
        return feature_tmp

    def show(self):
        """打印全部属性"""
        print(f"本次自定义过程清洗的特征：{self.cleaned_var}")
        print(f"本次自定义过程组合的特征：{self.used_var}")
        print(f"本次自定义过程调整的特征：{self.adjust_var}")

    def model(self):
        """
        自定义模型类
        """
        # ------------------------------------------
        # eg.
        #   from statsmodels.api import Logit
        #   return Logit
        # ------------------------------------------


if __name__ == "__main__":
    # 自定义调试用代码
    import mksc
    from mksc.feature_engineering import preprocess
    # 加载数据、变量类型划分、特征集与标签列划分
    data = mksc.load_data()
    numeric_var, category_var, datetime_var, label_var = preprocess.get_variable_type()
    feature = data[numeric_var + category_var + datetime_var]
    label = data[label_var]
    try:
        feature[numeric_var] = feature[numeric_var].astype('float')
        feature[category_var] = feature[category_var].astype('object')
        feature[datetime_var] = feature[datetime_var].astype('datetime64')
    except Exception as e:
        print(e)
